"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.inputStep = exports.graphqlGetTypeForNode = exports.assertInputStep = void 0;
const tslib_1 = require("tslib");
const graphql = tslib_1.__importStar(require("graphql"));
const inspect_js_1 = require("./inspect.js");
const __inputDynamicScalar_js_1 = require("./steps/__inputDynamicScalar.js");
const __inputObject_js_1 = require("./steps/__inputObject.js");
const index_js_1 = require("./steps/index.js");
const { assertScalarType, GraphQLError, GraphQLList, GraphQLNonNull, isInputType, isLeafType, isInputObjectType, Kind, } = graphql;
function assertInputStep(itemPlan) {
    if (itemPlan instanceof index_js_1.__TrackedValueStep)
        return;
    if (itemPlan instanceof index_js_1.__InputListStep)
        return;
    if (itemPlan instanceof index_js_1.__InputStaticLeafStep)
        return;
    if (itemPlan instanceof __inputObject_js_1.__InputObjectStep)
        return;
    throw new Error(`Expected an InputStep, but found ${itemPlan}`);
}
exports.assertInputStep = assertInputStep;
function graphqlGetTypeForNode(operationPlan, node) {
    switch (node.kind) {
        case Kind.NAMED_TYPE: {
            const type = operationPlan.schema.getType(node.name.value);
            if (!type) {
                // Should not happen since the GraphQL operation has already been
                // validated against the schema.
                throw new Error(`Could not find type with name '${node.name.value}' in the schema`);
            }
            return type;
        }
        case Kind.LIST_TYPE:
            return new GraphQLList(graphqlGetTypeForNode(operationPlan, node.type));
        case Kind.NON_NULL_TYPE:
            return new GraphQLNonNull(graphqlGetTypeForNode(operationPlan, node.type));
        default: {
            const never = node;
            throw new Error(`Unknown node kind; node: ${(0, inspect_js_1.inspect)(never)}`);
        }
    }
}
exports.graphqlGetTypeForNode = graphqlGetTypeForNode;
/**
 * Returns a plan for the given `rawInputValue` AST node which could be a
 * variable or a literal, and could be nested so that a variable (or more than
 * one) appears somewhere. More than one plan may be created.
 *
 * @internal
 */
function inputStep(operationPlan, inputType, rawInputValue, defaultValue = undefined) {
    // This prevents recursion
    if (rawInputValue === undefined && defaultValue === undefined) {
        return (0, index_js_1.constant)(undefined);
    }
    const isObj = isInputObjectType(inputType);
    let inputValue = rawInputValue;
    if (inputValue?.kind === "Variable") {
        const variableName = inputValue.name.value;
        const variableDefinition = operationPlan.operation.variableDefinitions?.find((def) => def.variable.name.value === variableName);
        if (!variableDefinition) {
            // Should not happen since the GraphQL operation has already been
            // validated.
            throw new Error(`No definition for variable '${variableName}' found`);
        }
        const variableType = graphqlGetTypeForNode(operationPlan, variableDefinition.type);
        if (!isInputType(variableType)) {
            throw new Error(`Expected varible type to be an input type`);
        }
        return inputVariablePlan(operationPlan, variableName, variableType, inputType, defaultValue);
    }
    // Note: past here we know whether `defaultValue` will be used or not because
    // we know `inputValue` is not a variable.
    inputValue = inputValue ?? defaultValue;
    if (inputType instanceof GraphQLNonNull) {
        const innerType = inputType.ofType;
        const valuePlan = inputStep(operationPlan, innerType, inputValue, undefined);
        return inputNonNullPlan(operationPlan, valuePlan);
    }
    else if (inputType instanceof GraphQLList) {
        return new index_js_1.__InputListStep(inputType, inputValue);
    }
    else if (isLeafType(inputType)) {
        if (inputValue?.kind === Kind.OBJECT || inputValue?.kind === Kind.LIST) {
            const scalarType = assertScalarType(inputType);
            return new __inputDynamicScalar_js_1.__InputDynamicScalarStep(scalarType, inputValue);
        }
        else {
            // Variable is already ruled out, so it must be one of: Kind.INT | Kind.FLOAT | Kind.STRING | Kind.BOOLEAN | Kind.NULL | Kind.ENUM
            // none of which can contain a variable:
            return new index_js_1.__InputStaticLeafStep(inputType, inputValue);
        }
    }
    else if (isObj) {
        return new __inputObject_js_1.__InputObjectStep(inputType, inputValue);
    }
    else {
        const never = inputType;
        throw new Error(`Unsupported type in inputPlan: '${(0, inspect_js_1.inspect)(never)}'`);
    }
}
exports.inputStep = inputStep;
function doTypesMatch(variableType, expectedType) {
    if (variableType instanceof GraphQLNonNull &&
        expectedType instanceof GraphQLNonNull) {
        return doTypesMatch(variableType.ofType, expectedType.ofType);
    }
    else if (variableType instanceof GraphQLNonNull) {
        // Variable is stricter than input type, that's fine
        return doTypesMatch(variableType.ofType, expectedType);
    }
    else if (variableType instanceof GraphQLList &&
        expectedType instanceof GraphQLList) {
        return doTypesMatch(variableType.ofType, expectedType.ofType);
    }
    else {
        return variableType === expectedType;
    }
}
function inputVariablePlan(operationPlan, variableName, variableType, inputType, defaultValue = undefined) {
    if (variableType instanceof GraphQLNonNull &&
        !(inputType instanceof GraphQLNonNull)) {
        const unwrappedVariableType = variableType.ofType;
        return inputVariablePlan(operationPlan, variableName, unwrappedVariableType, inputType, defaultValue);
    }
    const typesMatch = doTypesMatch(variableType, inputType);
    if (!typesMatch) {
        // REF: https://spec.graphql.org/draft/#IsVariableUsageAllowed()
        if (inputType instanceof GraphQLNonNull &&
            !(variableType instanceof GraphQLNonNull)) {
            const variablePlan = inputVariablePlan(operationPlan, variableName, variableType, inputType.ofType, defaultValue);
            if (variablePlan.evalIs(null) || variablePlan.evalIs(undefined)) {
                throw new GraphQLError(`Expected non-null value of type ${inputType.ofType.toString()}`);
            }
            return variablePlan;
        }
        throw new Error("Expected variable and input types to match");
    }
    const variableValuePlan = operationPlan.trackedVariableValuesStep.get(variableName);
    if (defaultValue === undefined || !variableValuePlan.evalIs(undefined)) {
        // There's no default value, or we know for sure that our variable will be
        // set (even if null) and thus the default will not be used; use the variable.
        return variableValuePlan;
    }
    else {
        // `defaultValue` is NOT undefined, and we know variableValue is
        // `undefined` (and always will be); we're going to loop back and pretend
        // that no value was passed in the first place (instead of the variable):
        return inputStep(operationPlan, inputType, undefined, defaultValue);
    }
}
/**
 * Implements `InputNonNullStep`.
 */
function inputNonNullPlan(_operationPlan, innerPlan) {
    return innerPlan;
}
//# sourceMappingURL=input.js.map