import type { GraphQLObjectType } from "graphql";
import type { LayerPlan, LayerPlanReasonSubroutine } from "./engine/LayerPlan.js";
import type { OperationPlan } from "./engine/OperationPlan.js";
import type { AddDependencyOptions, ExecutionDetails, ExecutionEntryFlags, GrafastResultsList, GrafastResultStreamList, JSONValue, PromiseOrDirect, StepOptimizeOptions, StreamDetails, UnbatchedExecutionExtra } from "./interfaces.js";
import { $$subroutine } from "./interfaces.js";
import type { __ItemStep } from "./steps/index.js";
declare function reallyAssertFinalized(plan: BaseStep): void;
export declare const assertFinalized: typeof reallyAssertFinalized;
/**
 * The base abstract plan type; you should not extend this directly - instead
 * use an ExecutableStep (for use when planning output fields) or a
 * ModifierStep (for use when planning arguments/input fields).
 *
 * @remarks
 *
 * Though it might seem that ModifierStep should be used for all inputs
 * (arguments, input objects), this is not the case. When planning an output
 * field, all inputs to it (including arguments, parent plan, context, etc)
 * should be other executable plans. The only time that ModifierStep is used is
 * when writing specific plans as part of the argument or input field
 * definitions themself; even in these cases the inputs to these plan resolvers
 * will be ExecutablePlans.
 */
export declare abstract class BaseStep {
    readonly operationPlan: OperationPlan;
    [$$subroutine]: LayerPlan<LayerPlanReasonSubroutine> | null;
    isArgumentsFinalized: boolean;
    isFinalized: boolean;
    debug: boolean;
    /**
     * Set this true for plans that implement mutations; this will prevent them
     * from being tree-shaken.
     */
    hasSideEffects: boolean;
    constructor();
    toString(): string;
    /**
     * This metadata will be merged into toString when referencing this plan.
     */
    toStringMeta(): string | null;
    planJSONExtra(): Record<string, JSONValue | undefined> | undefined;
    finalize(): void;
    destroy(): void;
}
/**
 * Executable plans are the plans associated with leaves on the GraphQL tree,
 * they must be able to execute to return values.
 */
export declare class ExecutableStep<TData = any> extends BaseStep {
    static $$export: any;
    /**
     * Setting this true is a performance optimisation, but it comes with strong
     * rules; we do not test you comply with these rules (as that would undo the
     * performance gains) but should you break them the behaviour is undefined
     * (and, basically, the schema may no longer be GraphQL compliant).
     *
     * Do not set this true unless the following hold:
     *
     * - The `execute` method must be a regular (not async) function
     * - The `execute` method must NEVER return a promise
     * - The values within the list returned from `execute` must NEVER include
     *   promises or FlaggedValue objects
     * - The result of calling `execute` should not differ after a
     *   `step.hasSideEffects` has executed (i.e. it should be pure, only
     *   dependent on its deps and use no external state)
     *
     * It's acceptable for the `execute` method to throw if it needs to.
     *
     * This optimisation applies to the majority of the built in plans and allows
     * the engine to execute without needing to resolve any promises which saves
     * precious event-loop ticks.
     */
    isSyncAndSafe: boolean;
    /**
     * (default = ALL_FLAGS & ~FLAG_NULL)
     */
    protected readonly defaultForbiddenFlags: ExecutionEntryFlags;
    /**
     * The plan this plan will need data from in order to execute.
     */
    protected readonly dependencies: ReadonlyArray<ExecutableStep>;
    /**
     * What execution entry flags we can't handle for the given indexed dependency
     * (default = this.defaultForbiddenFlags)
     */
    protected readonly dependencyForbiddenFlags: ReadonlyArray<ExecutionEntryFlags>;
    protected readonly dependencyOnReject: ReadonlyArray<Error | null | undefined>;
    /**
     * Unique identifier for this step within the plan.
     */
    readonly id: number;
    /**
    /**
     * True when `optimize` has been called at least once.
     */
    isOptimized: boolean;
    /**
     * Set this true if your plan's optimize method can be called a second time;
     * note that in this situation it's likely that your dependencies will not be
     * what you expect them to be (e.g. a PgSelectSingleStep might become an
     * AccessStep).
     */
    allowMultipleOptimizations: boolean;
    /**
     * Set the metaKey so `execute` will be passed a meta object to use.
     * Depending on what you set it to, you can share execution meta between
     * multiple steps of the same class (or even a family of step classes).
     *
     * A sensible value for it is `this.metaKey = this.id;`.
     */
    metaKey: number | string | symbol | undefined;
    /**
     * Like `metaKey` but for the optimize phase
     */
    optimizeMetaKey: number | string | symbol | undefined;
    constructor();
    protected withMyLayerPlan<T>(callback: () => T): T;
    protected getStep(id: number): ExecutableStep;
    protected getDepOptions(depId: number): AddDependencyOptions;
    protected getDep(_depId: number): ExecutableStep;
    /**
     * Like getDep, except it skips over __ItemStep and similar steps to get to
     * where the parent really is.
     *
     * @experimental
     */
    protected getDepDeep(depId: number): ExecutableStep;
    toString(): string;
    protected canAddDependency(step: ExecutableStep): boolean;
    protected addDependency(stepOrOptions: ExecutableStep | AddDependencyOptions): number;
    /**
     * Adds "unary" dependencies; in `execute({count, values})` you'll receive a
     * `values[index]` (where `index` is the return value of this function) with
     * `isBatch = false` so you can use the `values[index].value` property
     * directly.
     */
    protected addUnaryDependency(stepOrOptions: ExecutableStep | AddDependencyOptions): number;
    /**
     * Given a list of "peer" steps, return a list of these `peers` that are
     * equivalent to this step.
     *
     * NOTE: equivalence goes both ways: `a.deduplicate([b]).includes(b)` if and
     * only if `b.deduplicate([a]).includes(a)`.
     *
     * If you need to transform the peer to be equivalent you should do so via
     * the `deduplicatedWith` callback later.
     */
    deduplicate?(_peers: readonly ExecutableStep[]): readonly ExecutableStep[];
    /**
     * If this plan is replaced via deduplication, this method gives it a chance
     * to hand over its responsibilities to its replacement.
     */
    deduplicatedWith?(replacement: ExecutableStep): void;
    /**
     * Our chance to optimise the plan (which could go as far as to inline the
     * plan into the parent plan).
     */
    optimize?(_options: StepOptimizeOptions): ExecutableStep;
    finalize(): void;
    /**
     * This function will be called with 'execution details', an object containing:
     *
     * - `count`: the number of entries in the batch that's being executed
     * - `values`: a tuple representing the runtime values of the steps
     *   dependencies; each value in the tuple is an object, either a batch object
     *   containing a list of size `count` containing the values, or a unary
     *   object containing the single value common to all entries.
     * - `indexMap`: helper function to map over each index from `0` to `count-1`,
     *   returning the resulting array.
     * - `indexForEach`: as `indexMap`, but without the array result.
     * - `meta`: [experimental]
     *
     * `execute` must return a list with `count` entries, where each value in the
     * list relates to the result of executing this plan for the corresponding
     * entry in each of the entries in the `values` tuple.
     *
     * IMPORTANT: it is up to the execute function to cache/memoize results as
     * appropriate for performance, this can be done via the `meta` object.
     *
     * The `meta` object is an empty object stored to `grafastContext.metaByPlan`
     * that can be used to store anything this plan needs. We recommend that you
     * add attributes to meta for each purpose (e.g. use `meta.cache` for
     * memoizing results) so that you can expand your usage of meta in future.
     */
    execute(details: ExecutionDetails): PromiseOrDirect<GrafastResultsList<TData>>;
    destroy(): void;
}
export declare abstract class UnbatchedExecutableStep<TData = any> extends ExecutableStep<TData> {
    static $$export: {
        moduleName: string;
        exportName: string;
    };
    finalize(): void;
    execute({ indexMap, values, extra, }: ExecutionDetails): PromiseOrDirect<GrafastResultsList<TData>>;
    abstract unbatchedExecute(extra: UnbatchedExecutionExtra, ...tuple: any[]): PromiseOrDirect<TData>;
}
export declare function isExecutableStep<TData = any>(step: unknown): step is ExecutableStep<TData>;
export declare function assertExecutableStep<TData>(step: BaseStep): asserts step is ExecutableStep<TData>;
export declare function isUnbatchedExecutableStep<TData = any>(step: unknown): step is UnbatchedExecutableStep<TData>;
export type ObjectLikeStep<TData extends {
    [key: string]: ExecutableStep;
} = {
    [key: string]: ExecutableStep;
}> = ExecutableStep<{
    [key in keyof TData]: TData[key] extends ExecutableStep<infer U> ? U : never;
}> & {
    get<TKey extends keyof TData>(key: TKey): ExecutableStep<TData[TKey]>;
};
export declare function isObjectLikeStep<TData extends {
    [key: string]: ExecutableStep;
} = {
    [key: string]: ExecutableStep;
}>(plan: ExecutableStep): plan is ObjectLikeStep<TData>;
export type ListLikeStep<TData extends [...ExecutableStep[]] = [...ExecutableStep[]]> = ExecutableStep<{
    [key in keyof TData]: TData[key] extends ExecutableStep<infer U> ? U : never;
}> & {
    at<TKey extends keyof TData>(key: TKey): ExecutableStep<TData[TKey]>;
};
export declare function isListLikeStep<TData extends [...ExecutableStep[]] = [...ExecutableStep[]]>(plan: ExecutableStep): plan is ListLikeStep<TData>;
export type StreamableStep<TData> = ExecutableStep<ReadonlyArray<TData>> & {
    stream(details: StreamDetails): PromiseOrDirect<GrafastResultStreamList<TData>>;
};
export declare function isStreamableStep<TData>(plan: ExecutableStep<ReadonlyArray<TData>>): plan is StreamableStep<TData>;
export type PolymorphicStep = ExecutableStep & {
    planForType(objectType: GraphQLObjectType): ExecutableStep;
};
export declare function isPolymorphicStep(s: ExecutableStep): s is PolymorphicStep;
/**
 * Modifier plans modify their parent plan (which may be another ModifierStep
 * or an ExecutableStep). First they gather all the requirements from their
 * children (if any) being applied to them, then they apply themselves to their
 * parent plan. This application is done through the `apply()` method.
 *
 * Modifier plans do not use dependencies.
 */
export declare abstract class ModifierStep<TParentStep extends BaseStep = BaseStep> extends BaseStep {
    protected readonly $parent: TParentStep;
    static $$export: any;
    readonly id: string;
    constructor($parent: TParentStep);
    toString(): string;
    /**
     * In this method, you should apply the changes to your `this.$parent` plan
     */
    abstract apply(): void;
}
export declare function isModifierStep<TParentStep extends ExecutableStep | ModifierStep<any>>(plan: BaseStep): plan is ModifierStep<TParentStep>;
export declare function assertModifierStep<TParentStep extends ExecutableStep | ModifierStep<any>>(plan: BaseStep, pathDescription: string): asserts plan is ModifierStep<TParentStep>;
export interface ListCapableStep<TOutputData, TItemStep extends ExecutableStep<TOutputData> = ExecutableStep<TOutputData>> extends ExecutableStep<ReadonlyArray<any>> {
    listItem(itemPlan: __ItemStep<this>): TItemStep;
}
export declare function isListCapableStep<TData, TItemStep extends ExecutableStep<TData>>(plan: ExecutableStep<ReadonlyArray<TData>>): plan is ListCapableStep<TData, TItemStep>;
export declare function assertListCapableStep<TData, TItemStep extends ExecutableStep<TData>>(plan: ExecutableStep<ReadonlyArray<TData>>, pathDescription: string): asserts plan is ListCapableStep<TData, TItemStep>;
export {};
//# sourceMappingURL=step.d.ts.map